from __future__ import unicode_literals
'''
Inspired by http://www.thescripts.com/forum/thread46361.html

Derived from github.com/esss/ben10.
'''
import six
import locale



#===================================================================================================
# reraiseit
#===================================================================================================
def reraise(exception, message, separator='\n'):
    '''
    Raised the same exception given, with an additional message.

    :param Exception exception:
        Original exception being raised with additional messages

    :param unicode message:
        Message to be added to the given exception

    :param unicode separator:
        String separating `message` from the `exception`'s original message.

    e.g.
        try:
            raise RuntimeError('original message')
        except Exception, e:
            Reraise(e, 'message')

        >>> RuntimeError:
        >>> message
        >>> original message

        try:
            raise RuntimeError('original message')
        except Exception, e:
            Reraise(e, '[message]', separator=' ')

        >>> RuntimeError:
        >>> [message] original message
    '''
    import sys

    # IMPORTANT: Do NOT use try/except mechanisms in this method or the sys.exc_info()[-1] will be invalid

    if hasattr(exception, 'reraised_message'):
        current_message = exception.reraised_message
    else:
        current_message = exception_to_unicode(exception)

    # Build the new message
    if not current_message.startswith(separator):
        current_message = separator + current_message
    message = '\n' + message + current_message

    if exception.__class__ in _SPECIAL_EXCEPTION_MAP:
        # Handling for special case, some exceptions have different behaviors.
        exception = _SPECIAL_EXCEPTION_MAP[exception.__class__](*exception.args)

    elif exception.__class__ not in _SPECIAL_EXCEPTION_MAP.values():
        # In Python 2.5 overriding the exception "__str__" has no effect in "unicode()". Instead, we
        # must change the "args" attribute which is used to build the string representation.
        # Even though the documentation says "args" will be deprecated, it uses its first argument
        # in unicode() implementation and not "message".
        exception.args = (message,)

    exception.message = message
    # keep the already decoded message in the object in case this exception is reraised again
    exception.reraised_message = message

    # Reraise the exception with the EXTRA message information
    if six.PY2:
        six.reraise(exception, None, sys.exc_info()[-1])
    else:
        raise exception.with_traceback(sys.exc_info()[-1])


#===================================================================================================
# exception_to_unicode
#===================================================================================================
def exception_to_unicode(exception):
    '''
    Obtains unicode representation of an Exception.

    This wrapper is used to circumvent Python 2.7 problems with built-in exceptions with unicode
    messages.

    Steps used:
        * Try to obtain Exception.__unicode__
        * Try to obtain Exception.__str__ and decode with utf-8
        * Try to obtain Exception.__str__ and decode with locale.getpreferredencoding
        * If all fails, return Exception.__str__ and decode with (ascii, errors='replace')

    :param Exception exception:

    :return unicode:
        Unicode representation of an Exception.
    '''
    if six.PY2:
        try:
            # First, try to obtain __unicode__ as defined by the Exception
            return six.text_type(exception)
        except UnicodeDecodeError:
            try:
                # If that fails, try decoding with utf-8 which is the strictest and will complain loudly.
                return bytes(exception).decode('utf-8')
            except UnicodeDecodeError:
                try:
                    # If that fails, try obtaining bytes repr and decoding with locale
                    return bytes(exception).decode(locale.getpreferredencoding())
                except UnicodeDecodeError:
                    # If all failed, give up and decode with ascii replacing errors.
                    return bytes(exception).decode(errors='replace')
        except UnicodeEncodeError:
            # Some exception contain unicode messages, but try to convert them to bytes when calling
            # unicode() (such as IOError). In these cases, we do our best to fix Python 2.7's poor
            # handling of unicode exceptions.
            assert type(exception.message) == six.text_type  # This should be true if code got here.
            return exception.message
    else:
        return str(exception)


#===================================================================================================
# SPECIAL_EXCEPTIONS
#===================================================================================================
# [[[cog
# SPECIAL_EXCEPTIONS = [
#     KeyError,
#     OSError,
#     IOError,
#     SyntaxError,
#     UnicodeDecodeError,
#     UnicodeEncodeError,
# ]
# from ben10.foundation.string import Dedent
# exception_map = []
# for exception_class in SPECIAL_EXCEPTIONS:
#     superclass_name = exception_class.__name__
#     exception_map.append('\n        ' + superclass_name + ' : Reraised' + superclass_name + ',')
#     cog.out(Dedent(
#         '''
#         class Reraised%(superclass_name)s(%(superclass_name)s):
#             def __init__(self, *args):
#                 %(superclass_name)s.__init__(self, *args)
#                 self.message = None
#
#             def __str__(self):
#                 return self.message
#
#
#         '''% locals()
#     ))
# cog.out(Dedent(
#     '''
#     _SPECIAL_EXCEPTION_MAP = {%s
#     }
#     ''' % ''.join(exception_map)
# ))
# ]]]
class ReraisedKeyError(KeyError):
    def __init__(self, *args):
        KeyError.__init__(self, *args)
        self.message = None

    def __str__(self):
        return self.message

class ReraisedOSError(OSError):
    def __init__(self, *args):
        OSError.__init__(self, *args)
        self.message = None

    def __str__(self):
        return self.message

class ReraisedOSError(OSError):
    def __init__(self, *args):
        OSError.__init__(self, *args)
        self.message = None

    def __str__(self):
        return self.message

class ReraisedSyntaxError(SyntaxError):
    def __init__(self, *args):
        SyntaxError.__init__(self, *args)
        self.message = None

    def __str__(self):
        return self.message

class ReraisedUnicodeDecodeError(UnicodeDecodeError):
    def __init__(self, *args):
        UnicodeDecodeError.__init__(self, *args)
        self.message = None

    def __str__(self):
        return self.message

class ReraisedUnicodeEncodeError(UnicodeEncodeError):
    def __init__(self, *args):
        UnicodeEncodeError.__init__(self, *args)
        self.message = None

    def __str__(self):
        return self.message

_SPECIAL_EXCEPTION_MAP = {
    KeyError : ReraisedKeyError,
    OSError : ReraisedOSError,
    OSError : ReraisedOSError,
    SyntaxError : ReraisedSyntaxError,
    UnicodeDecodeError : ReraisedUnicodeDecodeError,
    UnicodeEncodeError : ReraisedUnicodeEncodeError,
}
# [[[end]]] (checksum: 896c3faa794c9a17cbe89209d38816dc)


if six.PY3:
    class ReraisedFileNotFoundError(FileNotFoundError):
        def __init__(self, *args):
            FileNotFoundError.__init__(self, *args)
            self.message = None

        def __str__(self):
            return self.message

    _SPECIAL_EXCEPTION_MAP[FileNotFoundError] = ReraisedFileNotFoundError



# coding=utf-8
from __future__ import unicode_literals, print_function

import locale  # Used by code inside a string.
import pytest
import six
import traceback

from zerotk.reraiseit import reraise, exception_to_unicode

MESSAGE_TYPE = bytes if six.PY2 else str

#===================================================================================================
# ExecutePythonCode
#===================================================================================================
def execute_python_code(code):
    try:
        c = compile(code, '<string>', 'exec')
    except Exception as e:
        print('On execute_python_code with code=\n{}'.format(code))
        raise
    six.exec_(c)


#===================================================================================================
# MessageAttributeIsBytesException
#===================================================================================================
class MessageAttributeIsBytesException(Exception):
    '''
    Simulate a 'message' attribute with the bytes type

    :ivar bytes message:
    '''
    def __init__(self):
        self.message = b'My ascii message'

    def __str__(self):
        return bytes('исключение')


#===================================================================================================
# ExceptionTestConfiguration
#===================================================================================================
class ExceptionTestConfiguration():

    def __init__(self, exception_type, string_statement, expected_inner_exception_message='', expected_traceback_message=None, raised_exception=None):
        self.exception_type = exception_type
        if raised_exception is not None:
            self.raised_exception = raised_exception
        else:
            self.raised_exception = exception_type
        self.string_statement = string_statement
        self.expected_inner_exception_message = expected_inner_exception_message
        self.expected_traceback_message = expected_traceback_message


    def RaiseExceptionUsingReraise(self):

        def raise_exception():
            execute_python_code(self.string_statement)
            pytest.fail('Should not reach here')

        def reraise_exception():
            try:
                raise_exception()
            except self.raised_exception as e:
                reraise(e, "While doing 'bar'")

        try:
            try:
                reraise_exception()
            except self.raised_exception as e1:
                reraise(e1, "While doing x:")
        except self.raised_exception as e2:
            reraise(e2, "While doing y:")


    def get_expected_exception_message(self):
        return "\nWhile doing y:\nWhile doing x:\nWhile doing 'bar'\n" + self.expected_inner_exception_message


    def get_expected_traceback_message(self, actual_exception):
        # Getting the type of the "actual exception" because its type might be different than self.exception_type
        reraised_exception_name = type(actual_exception).__name__
        if six.PY3 and reraised_exception_name.startswith('Reraised'):
            reraised_exception_name = 'zerotk.reraiseit._reraiseit.' + reraised_exception_name
        exception_message = self.get_expected_exception_message()

        if self.expected_traceback_message is not None:
            exception_message = self.expected_traceback_message
        else:
            exception_message = reraised_exception_name + ": " + exception_message + "\n"

        # HACK [muenz]: the 'traceback' module does this, so in order to be able to compare strings we need this workaround
        # notice that if the exception "leaks" the Python console handles the unicode symbols properly
        exception_message = traceback._some_str(exception_message)
        assert type(exception_message) == MESSAGE_TYPE

        return exception_message


if six.PY2:
    parametrized_exceptions = pytest.mark.parametrize(
        'exception_configuration',
        [
            ExceptionTestConfiguration(ValueError, "raise ValueError('message')", 'message'),
            ExceptionTestConfiguration(KeyError, "raise KeyError('message')", "u'message'"),
            ExceptionTestConfiguration(OSError, "raise OSError(2, 'message')", '[Errno 2] message'),
            ExceptionTestConfiguration(IOError, "raise IOError('message')", 'message'),
            ExceptionTestConfiguration(
                SyntaxError,
                "in valid syntax",
                expected_inner_exception_message='invalid syntax (<string>, line 1)',
                expected_traceback_message=(
                    '  File "<string>", line 1\n'
                    '    in valid syntax\n'
                    '     ^\n'
                    'ReraisedSyntaxError: invalid syntax\n'
                )
            ),
            ExceptionTestConfiguration(
                UnicodeDecodeError,
                "u'£'.encode('utf-8').decode('ascii')",
                "'ascii' codec can't decode byte 0xc2 in position 0: ordinal not in range(128)"
            ),
            ExceptionTestConfiguration(
                UnicodeEncodeError,
                "u'£'.encode('ascii')",
                "'ascii' codec can't encode character u'\\xa3' in position 0: ordinal not in range(128)"
            ),
            ExceptionTestConfiguration(AttributeError, "raise AttributeError('message')", 'message'),

            ExceptionTestConfiguration(OSError, "raise OSError()"),
            ExceptionTestConfiguration(OSError, "raise OSError(1)", '1'),
            ExceptionTestConfiguration(OSError, "raise OSError(2, '£ message')", '[Errno 2] £ message'),
            ExceptionTestConfiguration(IOError, "raise IOError('исключение')", "исключение", expected_traceback_message='IOError: <unprintable IOError object>\n'),
            pytest.mark.xfail(raises=AssertionError, reason="ExceptionToUnicode() assumes that the 'message' attribute is 'unicode'")(
               ExceptionTestConfiguration(MessageAttributeIsBytesException, 'raise MessageAttributeIsBytesException()', '')
            ),
            ExceptionTestConfiguration(OSError, "raise OSError(2, '£ message'.encode(locale.getpreferredencoding()))", '[Errno 2] £ message'),
            ExceptionTestConfiguration(OSError, "raise OSError(2, b'£ message')", '[Errno 2] £ message'),
            ExceptionTestConfiguration(IOError, "raise IOError(b'£ message')", '£ message', expected_traceback_message='IOError: <unprintable IOError object>\n'),
            ExceptionTestConfiguration(Exception, "raise Exception(b'£ message')", '£ message'),
        ], ids=[
            'ValueError',
            'KeyError',
            'OSError',
            'IOError',
            'SyntaxError',
            'UnicodeDecodeError',
            'UnicodeEncodeError',
            'AttributeError',

            'OSError - empty',
            'OSError - ErrorNo, empty message',
            'OSError - ErrorNo, unicode message',
            'IOError - unicode message',
            'MessageAttributeIsBytesException',

            'OSError - bytes message in UTF-8',
            'OSError - bytes message in locale.getpreferredencoding()',
            'IOError - bytes message',
            'Exception - bytes message',
        ]
    )
else:
    parametrized_exceptions = pytest.mark.parametrize(
        'exception_configuration',
        [
            ExceptionTestConfiguration(ValueError, "raise ValueError('message')", 'message'),
            ExceptionTestConfiguration(KeyError, "raise KeyError('message')", "'message'"),
            ExceptionTestConfiguration(
                OSError,
                "raise OSError(2, 'message')",
                '[Errno 2] message',
            ),
            ExceptionTestConfiguration(IOError, "raise IOError('message')", 'message'),
            ExceptionTestConfiguration(
                SyntaxError,
                "in valid syntax",
                expected_inner_exception_message='invalid syntax (<string>, line 1)',
                expected_traceback_message=(
                    '  File "<string>", line 1\n'
                    '    in valid syntax\n'
                    '     ^\n'
                    'zerotk.reraiseit._reraiseit.ReraisedSyntaxError: invalid syntax\n'
                )
            ),
            ExceptionTestConfiguration(
                UnicodeDecodeError,
                "'£'.encode('utf-8').decode('ascii')",
                "'ascii' codec can't decode byte 0xc2 in position 0: ordinal not in range(128)"
            ),
            ExceptionTestConfiguration(
                UnicodeEncodeError,
                "'£'.encode('ascii')",
                "'ascii' codec can't encode character '\\xa3' in position 0: ordinal not in range(128)"
            ),
            ExceptionTestConfiguration(AttributeError, "raise AttributeError('message')", 'message'),

            # On Python 3 (Windows) raising OSError actually raises a FileNotFoundError.
            ExceptionTestConfiguration(OSError, "raise OSError()"),
            ExceptionTestConfiguration(OSError, "raise OSError(1)", '1'),
        ], ids=[
            'ValueError',
            'KeyError',
            'OSError',
            'IOError',
            'SyntaxError',
            'UnicodeDecodeError',
            'UnicodeEncodeError',
            'AttributeError',

            'OSError - empty',
            'OSError - ErrorNo, empty message',
        ]
    )


@parametrized_exceptions
def testReraiseKeepsTraceback(exception_configuration):
    with pytest.raises(exception_configuration.exception_type) as e:
        exception_configuration.RaiseExceptionUsingReraise()

    # reraiseit() should not appear in the traceback
    traceback_filenames = []
    for traceback_entry in e.traceback:
        try:
            traceback_filenames.append(traceback_entry.path.basename)
        except AttributeError:
            # traceback_entry.path will be a string when the code was run using "exec compile('code')".
            # However this does not hold true if 'code' contains a SyntaxError, which in that case 'traceback_entry.path'
            # will be a LocalPath
            traceback_filenames.append(traceback_entry.path)

    crash_entry = e.traceback.getcrashentry()
    expected_filenames = [
        'reraiseit_test.py',
        'reraiseit_test.py',
        '_reraiseit.py',
        'reraiseit_test.py',
        '_reraiseit.py',
        'reraiseit_test.py',
        'reraiseit_test.py',
        '_reraiseit.py',
        'reraiseit_test.py',
        'reraiseit_test.py',
        'reraiseit_test.py',
    ]
    if isinstance(crash_entry.path, MESSAGE_TYPE):
        if six.PY2:
            expected_filenames += ['six.py', '<string>']
        expected_filenames += ['<string>']
    if six.PY2:
        expected_filenames = map(six.binary_type, expected_filenames)
    assert traceback_filenames == expected_filenames

    assert crash_entry.locals['code'] == exception_configuration.string_statement


@parametrized_exceptions
def testReraiseAddsMessagesCorrectly(exception_configuration):
    with pytest.raises(exception_configuration.exception_type) as e:
        exception_configuration.RaiseExceptionUsingReraise()

    assert isinstance(e.value, exception_configuration.exception_type)
    assert exception_to_unicode(e.value) == exception_configuration.get_expected_exception_message()

    traceback_message = traceback.format_exception_only(type(e.value), e.value)
    traceback_message = ''.join(traceback_message)

    assert traceback_message == exception_configuration.get_expected_traceback_message(e.value)


@parametrized_exceptions
def testPickle(exception_configuration):
    try:
        exception_configuration.RaiseExceptionUsingReraise()
    except exception_configuration.exception_type as reraised_exception:
        from six.moves.cPickle import dumps, loads
        dumped_exception = dumps(reraised_exception)
        pickled_exception = loads(dumped_exception)
        assert exception_to_unicode(pickled_exception) == exception_to_unicode(reraised_exception)
        assert exception_to_unicode(pickled_exception) != ''
        assert exception_to_unicode(reraised_exception) != ''



#!/bin/env python
from setuptools import setup


setup(
    name='zerotk.reraiseit',
    version='2.0.2',

    author='Alexandre Andrade',
    author_email='kaniabi@gmail.com',

    url='https://github.com/zerotk/reraiseit',

    description = 'Reraise exceptions.',
    long_description = '''A function to re-raise exceptions adding information to the traceback and with unicode support.''',

    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 5 - Production/Stable',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
    ],

    include_package_data=True,

    packages=['zerotk', 'zerotk.reraiseit'],
    namespace_packages=['zerotk'],

    install_requires=[
        'six',
    ],
    setup_requires=[
        'pytest_runner',
    ],
    tests_require=[
        'cogapp',
        'coverage',
        'pytest',
    ],
)
